;! /* 
;!  * A self-stabilizing algorithm simulator by ring of PICs
;!  * Hirotsugu Kakugawa
;!  *
;!  *
;!  * Edition History
;!  *  17 Jul 2002  Code Outline
;!  *  23 Jul 2002  Implemented MUTEX and MIS.
;!  *  29 Jul 2002  Debugged MUTEX.
;!  */			
;!  


	list
	radix	dec
	
;;;	
;;; Target Processor: 16F84
;;;	
	include		pic16f84.h


;;;
;;; Configuration:  CP=off, DP=off, PWRTE=on, WDTE=off, FOSC=HS
	__config	_CP_OFF | _DP_OFF | _PWRTE_ON | _WDTE_OFF | _OSC_HS


;! /* Algorithm Type */
ALG_MUTEX	EQU		0
ALG_MIS		EQU		1
ALG_LE		EQU		2
ALG_TEST	EQU		3


	
;! /* PIC 16F84A I/O Pin Assignments
;!  *
;!  * Common:
;!  *    - PA0       : (in)  Algorithm select #0
;!  *    - PA1       : (in)  Algorithm select #1
;!  *    - PA2       : (in)  Transient fault
;!  *    - PA3       : (out) LED
;!  *
;!  * Unidirectional Ring:
;!  *    - PB0...PB3 : (in)   Data in   (from left)
;!  *    - PB4...PB7 : (out)  Data out  (to right)
;!  *
;!  * Unidirectional Ring:
;!  *    - PB0, PB1 : (out)   Data in   (to   left)
;!  *    - PB2, PB3 : (in)    Data out  (from left)
;!  *    - PB4, PB5 : (in)    Data out  (to right)
;!  *    - PB6, PB7 : (out)   Data in   (from right)
;!  */

PA_SELECT0		EQU		0
PA_SELECT1		EQU		1
PA_FAULT		EQU		2
PA_LED			EQU		3

;; Bidirectional Ring
;; 
;;              +-----------------+
;;       -----> | RB0         RB4 |------>   
;;       -----> | RB1         RB5 |------> 
;; LEFT         |                 |          RIGHT
;;       <----- | RB2         RB6 |<-----
;;       <----- | RB3         RB7 |<-----
;;              |                 |
;;              |   PIC 16F84A    |
;;              +-----------------+


;; Unidirectional Ring
;; 
;; LEFT         +-----------------+          RIGHT
;;       -----> | RB0         RB4 |------>   
;;       -----> | RB1         RB5 |------> 
;;       -----> | RB2         RB6 |------> 
;;       -----> | RB3         RB7 |------> 
;;              |                 |
;;              |   PIC 16F84A    |
;;              +-----------------+
	

	
		PAGE
;;; 
;;; Variables
;;; 
		DATA	


alg_type		EQU		0x10
rand			EQU		0x11
t			EQU		0x12
buff			EQU		0x13
buff2			EQU		0x14

;; MIS
l_mis_x			EQU		0x18
l_mis_l			EQU		0x19
l_mis_r			EQU		0x1a

;; MUTEX
l_mutex_x		EQU		0x18
l_mutex_y		EQU		0x19

;; misc
l_swa_t			EQU		0x1b
	
	
		PAGE
;;; 
;;; Program Text
;;; 
		ORG	0x0000
__crt0: 	GOTO	main

		ORG	0x0004
interrupt:	SLEEP

	
;! /*-----------------------------------------------*/ 
		ORG	0x0008
;! int  main()
main:
;! {


;!   init_io_mode();
	CALL	init_io_mode

;!   led_on(); delay_100ms(); led_off();
	CALL	led_on
	CALL	delay_100ms
	CALL	led_off
	
;!   test();		/* XXX  TEST  XXX */
;	CALL	test

;!   alg_type = sw_alg_type();
	CALL	sw_alg_type
	MOVWF	alg_type
		
;!   /* main loop */
;!   for (;;){
__main_0:
	
;!     if (alg_type != sw_alg_type()){
	CALL	sw_alg_type
	SUBWF	alg_type, W
	BTFSC	STATUS, Z
        GOTO    __main_2
;!       delay_100ms(); delay_100ms();
	CALL	delay_100ms
	CALL	delay_100ms
;!       if (alg_type != sw_alg_type()){
	CALL	sw_alg_type
	SUBWF	alg_type, W
	BTFSC	STATUS, Z
        GOTO    __main_1
;!         alg_type = sw_alg_type();
	CALL	sw_alg_type
	MOVWF	alg_type
;!         delay_100ms(); delay_100ms();
	CALL	delay_100ms
	CALL	delay_100ms
;!         continue;
	GOTO	__main_0
;!       }
__main_1:
;!     }
__main_2:	

;!     if (alg_type == ALG_MUTEX){
	MOVLW	ALG_MUTEX
	SUBWF	alg_type, W
	BTFSS	STATUS, Z
        GOTO    __main_3
;!     alg_switched();
	CALL	alg_switched
;!       ss_mutex();
	CALL	ss_mutex
	
;!     } else if (alg_type == ALG_MIS){
	GOTO	__main_5
__main_3: 
	MOVLW	ALG_MIS
	SUBWF	alg_type, W
	BTFSS	STATUS, Z
        GOTO    __main_4
;!     alg_switched();
	CALL	alg_switched
	CALL	alg_switched
;!       ss_mis();
	CALL	ss_mis
	
;!     } else {
	GOTO	__main_5
__main_4:
;!     alg_switched();
	CALL	alg_switched
	CALL	alg_switched
	CALL	alg_switched
;!       ss_test();
	CALL	ss_test
		
;!     } 
__main_5:
	
;!   }
	GOTO	__main_0

;!   /* NOTREACHED */
;! }


	
;! /*-----------------------------------------------*/ 
;! void  init_io_mode(void)
init_io_mode:
;! {
;!   port_a_io_mode(0x07);   /* Port A bit0,1,2: input */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0x07
	MOVWF	TRISA
	BCF	STATUS, RP0     ; select bank 0
;!   port_b_io_mode(0x00);   /* Port B all bits: output */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0x00
	MOVWF	TRISB
	BCF	STATUS, RP0     ; select bank 0
;! }
	RETURN


	

;! /*-----------------------------------------------*/ 
;! void  test(void)
test: 
;! {

;!   for (;;){
__test_0:

;!   led_on(); 
	CALL	led_on
	
;!     if (sw_fault() != 0){
	CALL	sw_fault
	BTFSC	STATUS, Z
	GOTO	__test_1
;!     delay_100ms(); 
	CALL	delay_100ms
	GOTO	__test_2
;!     } else {
__test_1: 
;!     delay_500ms(); 
	CALL	delay_500ms
;!     } 
__test_2: 

;!   led_off();
	CALL	led_off

;!     if (sw_fault() != 0){
	CALL	sw_fault
	BTFSC	STATUS, Z
	GOTO	__test_3
;!     delay_100ms(); 
	CALL	delay_100ms
	GOTO	__test_4
;!     } else {
__test_3: 
;!     delay_500ms(); 
	CALL	delay_500ms
;!     } 
__test_4: 

;!   }
	GOTO	__test_0
;!   /* NOTREACHED */
;! }
	RETURN	




	
		
;! /*-----------------------------------------------*/
;! /* SS Mutex
;!  *   *[   (x != (x_l+1) mod M)
;!  *	        ->  x := (x_l + 1) mod M;
;!  *    ]
;!  *   (M = 16 (0x0f))
;!  */
	
;! void  ss_mutex(void)
ss_mutex:	
;! {

;!   unidir_init_io();
	CALL	unidir_init_io
	
;!   delay_1s();
	CALL	delay_1s
	
;!   alg_type = ALG_MUTEX;
	MOVLW	ALG_MUTEX
	MOVWF	alg_type
	
;!   l_mutex_x = l_mutex_x & 0x0f;
	MOVF	l_mutex_x, W
	ANDLW	0x0f
	MOVWF	l_mutex_x
	
;!   for (;;){
__ss_mutex_0:	
		
;!     rand++;
	INCF	rand, F

;!     if (alg_type != sw_alg_type()){
	CALL	sw_alg_type
	SUBWF	alg_type, W
	BTFSC	STATUS, Z
        GOTO    __ss_mutex_01
;!       break;
        GOTO    __ss_mutex_6
;!     }
__ss_mutex_01:
	
;!     if (sw_fault() != 0){
	CALL	sw_fault
	BTFSC	STATUS, Z
	GOTO	__ss_mutex_1
;!       ss_mutex_fault();
	CALL	ss_mutex_fault
;!       unidir_write(l_mutex_x);
	MOVF	l_mutex_x, W
	CALL	unidir_write
;!       continue;
	GOTO	__ss_mutex_0
;!     }
__ss_mutex_1:
	
;!     unidir_write(l_mutex_x);
	MOVF	l_mutex_x, W
	CALL	unidir_write
	
;!     l_mutex_y = ss_mutex_left();
	CALL	ss_mutex_left
	MOVWF	l_mutex_y

;!     if (l_mutex_x != l_mutex_y){
	MOVF	l_mutex_y, W
	SUBWF	l_mutex_x, W
	BTFSC	STATUS, Z
        GOTO    __ss_mutex_3
	
;!       led_on();
	CALL	led_on
	
;!       delay_1ms(); delay_1ms(); 
	CALL	delay_1ms
	CALL	delay_1ms
	
;!       if (t == 0){
	MOVF	t, W
	BTFSS	STATUS, Z
	GOTO	__ss_mutex_2
;!         l_mutex_x = l_mutex_y;
	MOVF	l_mutex_y, W
	MOVWF	l_mutex_x	
;!         led_off();
	CALL	led_off
;!         t = delay_time();
	CALL	delay_time
	MOVWF	t
;!       }
__ss_mutex_2:
	
;!       t--;
	DECF	t, F
	
;!     } else {
	GOTO	__ss_mutex_4
__ss_mutex_3:	
;!       led_off();
	CALL	led_off
;!       t = delay_time();
	CALL	delay_time
	MOVWF	t
;!     }
__ss_mutex_4:
	
;!   }
	GOTO	__ss_mutex_0
;! }
__ss_mutex_6:
	RETURN	
	
;! void ss_mutex_left(void)
ss_mutex_left:		
;! { 
;!     l_mutex_y = unidir_read();
	CALL	unidir_read
	MOVWF	l_mutex_y
;!     l_mutex_y = (l_mutex_y + 1) % M;
	INCF	l_mutex_y, W
	ANDLW	0x0f
	MOVWF	l_mutex_y
;! }
	RETURN

;! void ss_mutex_fault(void)
ss_mutex_fault:		
;! { 
;!     l_mutex_x = (l_mutex_x + 1) % M;
	INCF	l_mutex_x, W
	ANDLW	0x0f
	MOVWF	l_mutex_x
;!     unidir_write(l_mutex_x);
	MOVF	l_mutex_x, W
	CALL	unidir_write
;! } 
	RETURN


	
;! /*-----------------------------------------------*/
;! /* SS MIS
;!  *   *[   ((x == 0) && (x_l == 0) && (x_r == 0))
;!  *	        ->  x := 1
;!  *     [] ((x == 1) && ((x_l != 0) || (x_r != 0)))
;!  *	        ->  x := 0
;!  *    ]	
;!  */

;! void ss_mis(void)
ss_mis:		
;! {
	
;!   bidir_init_io();
	CALL	bidir_init_io
	
;!   delay_1s();
	CALL	delay_1s
	
;!   alg_type = ALG_MIS;
	MOVLW	ALG_MIS
	MOVWF	alg_type

;!   l_mis_x = l_mis_x & 0x01;
	MOVF	l_mis_x, W
	ANDLW	0x01
	MOVWF	l_mis_x
	
;!   for (;;){
__ss_mis_0:
	
;!     if (alg_type != sw_alg_type()){
	CALL	sw_alg_type
	SUBWF	alg_type, W
	BTFSC	STATUS, Z
        GOTO    __ss_mis_9
;!       break;
        GOTO    __ss_mis_10
;!     }
__ss_mis_9:
		
;!     rand++;
	INCF	rand, F
	
;!     if (sw_fault() != 0){
	CALL	sw_fault
	BTFSC	STATUS, Z
	GOTO	__ss_mis_1
;!       ss_mis_fault();
	CALL	ss_mis_fault
;!       continue;
	GOTO	__ss_mis_0
;!     }
	
__ss_mis_1:	
;!     l_mis_l = bidir_read_l();
	CALL	bidir_read_l
	MOVWF	l_mis_l
;!     l_mis_r = bidir_read_r();
	CALL	bidir_read_r
	MOVWF	l_mis_r
	
;!     if ((l_mis_x == 0) && (l_mis_l == 0) && (l_mis_r == 0)){
	MOVF	l_mis_x, W	; if (l_mis_x != 0) goto __ss_mis_3
	BTFSS	STATUS, Z
	GOTO	__ss_mis_3
	MOVF	l_mis_l, W	; if (l_mis_l != 0) goto __ss_mis_3
	BTFSS	STATUS, Z
	GOTO	__ss_mis_3
	MOVF	l_mis_r, W	; if (l_mis_r != 0) goto __ss_mis_3
	BTFSS	STATUS, Z
	GOTO	__ss_mis_3
;!       delay_10ms(); 
	CALL	delay_10ms
;!       if (t == 0){
	MOVF	t, W
	BTFSS	STATUS, Z
	GOTO	__ss_mis_2
;!         l_mis_x = 1;
	MOVLW	1
	MOVWF	l_mis_x
;!         t = delay_time();
	CALL	delay_time
	MOVWF	t
;!       }
__ss_mis_2:	
;!       t--;
	DECF	t, F	
	
;!     } else if ((l_mis_x != 0) && ((l_mis_l != 0) || (l_mis_r != 0))){
	GOTO	__ss_mis_6
__ss_mis_3:	
	MOVF	l_mis_x, W	;   if (l_mis_x == 0) goto __ss_mis_5
	BTFSC	STATUS, Z
        GOTO    __ss_mis_5
	MOVF	l_mis_l, W	;   if (l_mis_l != 0) goto __ss_mis_3_1 
	BTFSS	STATUS, Z
	GOTO	__ss_mis_3_1
	MOVF	l_mis_r, W	;   if (l_mis_r != 0) goto __ss_mis_3_1 
	BTFSS	STATUS, Z
	GOTO	__ss_mis_3_1
	GOTO	__ss_mis_5	;   goto __ss_mis_5
__ss_mis_3_1:			; __ss_mis_3_1:	
;!       delay_10ms();
	CALL	delay_10ms
;!       if (t == 0){
	MOVF	t, W
	BTFSS	STATUS, Z
	GOTO	__ss_mis_4
;!         l_mis_x = 0;
	MOVLW	0
	MOVWF	l_mis_x
;!         t = delay_time();
	CALL	delay_time
	MOVWF	t
;!       }
__ss_mis_4: 
;!       t--;
	DECF	t, F
	
;!     } else {
	GOTO	__ss_mis_6
__ss_mis_5:	
;!       t = delay_time();
	CALL	delay_time
	MOVWF	t

;!     }
__ss_mis_6:
	
;!     bidir_write(l_mis_x);
	MOVF	l_mis_x, W
	CALL	bidir_write
;!     if (l_mis_x == 0){
	MOVF	l_mis_x, W
	BTFSS	STATUS, Z
        GOTO    __ss_mis_7
;!       led_off();
	CALL	led_off
;!     } else {
	GOTO	__ss_mis_8
__ss_mis_7: 
;!       led_on();
	CALL	led_on
;!     }
__ss_mis_8:
	
;!   }
	GOTO	__ss_mis_0
__ss_mis_10:
	
;! }
	RETURN
	

;! void ss_mis_fault(void)
ss_mis_fault:		
;! { 
;!     l_mis_x = (l_mis_x + 1) % 2;
	INCF	l_mis_x, W
	ANDLW	0x01
	MOVWF	l_mis_x
;!     unidir_write(l_mis_x);
	MOVF	l_mis_x, W
	CALL	bidir_write
;! } 
	RETURN

	
;! /*-----------------------------------------------*/
;! /* TEST
;!  *
;!  * LED On : Short   - Self-Diagnosis Error
;!  * LED On : Long    - Self-Diagnosis Ok
;!  * LED Off: Short   - Fault SW On
;!  * LED Off: Long    - Fault SW Off
;!  */

l_test_tmp	EQU		0x18

;! void ss_test(void)
ss_test: 
;! {

;!   unidir_init_io();
	CALL	unidir_init_io
	
;!   unidir_write(0x01);
	MOVLW	0x01
	CALL	unidir_write
	
;!   delay_1s();
	CALL	delay_1s
	
;!   alg_type = ALG_TEST;
	MOVLW	ALG_TEST
	MOVWF	alg_type 
	
;!   for (;;){
__ss_test_0:
	
;!     unidir_write(0x01);
	MOVLW	0x01
	CALL	unidir_write
	
;!     led_on();
	CALL	LED_ON
	
;!     if (unidir_read() != 0x01){
	CALL	unidir_read
	MOVWF	l_test_tmp
	MOVLW	0x01
	SUBWF	l_test_tmp
	BTFSC	STATUS, Z
	GOTO	__ss_test_1
;!     delay_100ms(); 
	CALL	delay_100ms
	GOTO	__ss_test_2
;!     } else {
__ss_test_1: 
;!     delay_1s(); 
	CALL	delay_1s
;!     } 
__ss_test_2: 

;!   led_off();
	CALL	led_off

;!     if (sw_fault() != 0){
	CALL	sw_fault
	BTFSC	STATUS, Z
	GOTO	__ss_test_3
;!     delay_100ms(); 
	CALL	delay_100ms
	GOTO	__ss_test_4
;!     } else {
__ss_test_3: 
;!     delay_1s(); 
	CALL	delay_1s
;!     } 
__ss_test_4: 

;!     if (alg_type != sw_alg_type()){
	CALL	sw_alg_type
	SUBWF	alg_type, W
	BTFSC	STATUS, Z
        GOTO    __ss_test_5
	 
;!       break;
	GOTO	__ss_test_6

;!     }
__ss_test_5:
	
;!   }
	GOTO	__ss_test_0
__ss_test_6:
	
;! }
	RETURN
			
	

;! /*-----------------------------------------------*/ 		
;! void delay_time(void)
delay_time:		
;! {
UNIT_TIME		EQU		 64

;!   rand &= 0x7f;
	MOVLW	0x7f
	ANDWF	rand, F
;!   return UNIT_TIME + rand;
	MOVLW	UNIT_TIME
	ADDWF	rand, W
	RETURN
;! }
	


	
;! /*-----------------------------------------------*/ 		
;! void unidir_init_io(void)
unidir_init_io:		
;! {
;!   port_a_io_mode(0x07);   /* Port A bit0,1,2: input */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0x07
	MOVWF	TRISA
	BCF	STATUS, RP0     ; select bank 0
;!   port_b_io_mode(0x0f);   /* Port B bit4,5,6,7 : input (00001111) */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0x0f
	MOVWF	TRISB
	BCF	STATUS, RP0     ; select bank 0
;! }
	RETURN
	
;! int unidir_read(void)
unidir_read:		
;! {
;!   buff = *port_b;
	MOVF	PORTB, W
	MOVWF	buff
;!   return buff & 0x0f;
	MOVLW	0x0F
	ANDWF	buff, W
;! }
	RETURN 
	
;! void unidir_write(int v)
unidir_write:	
;! {
;!   buff = v;
	MOVWF	buff
;!   buff = swap4bits(buff);
	SWAPF	buff, F
;!   buff &= 0xf0;
	MOVLW	0xF0
	ANDWF	buff, F
;!   *port_b = buff;
	MOVF	buff, W
	MOVWF	PORTB
;! }
	RETURN


	
;! /*-----------------------------------------------*/ 
;! void bidir_init_io(void)
bidir_init_io:	
;! {
;!   port_a_io_mode(0x07);   /* Port A bit0,1,2: input */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0x07
	MOVWF	TRISA
	BCF	STATUS, RP0     ; select bank 0
;!   port_b_io_mode(0xc3);   /* Port B bit 2,3,4,5 : input (11000011) */
	BSF	STATUS, RP0     ; select bank 1
	MOVLW	0xc3
	MOVWF	TRISB
	BCF	STATUS, RP0     ; select bank 0
;! }
	RETURN
	
;! int bidir_read_r(void)
bidir_read_r:
;! {
;!   buff = *port_b;
	MOVF	PORTB, W
	MOVWF	buff
;!   buff >>= 6;
	RRF	buff, F
	RRF	buff, F
	RRF	buff, F
	RRF	buff, F
	RRF	buff, F
	RRF	buff, F
;!   return buff & 0x03;
	MOVLW	0x03
	ANDWF	buff, W
	RETURN 
;! }
	
;! int bidir_read_l(void)
bidir_read_l:		
;! {
;!   buff = *port_b;
	MOVF	PORTB, W
	MOVWF	buff
;!   return buff & 0x03;
	MOVLW	0x03
	ANDWF	buff, W
	RETURN 
;! }

;! void bidir_write(int v)
bidir_write:		
;! {
;!   buff = v & 0x03;
	MOVWF	buff
	MOVLW	0x03
	ANDWF	buff, F
;!   buff2 = buff << 2;
	MOVF	buff, W
	MOVWF	buff2
	RLF	buff2, F
	RLF	buff2, F
;!   buff2 |= buff;
	MOVF	buff, W
	IORWF	buff2, F
;!   buff2 <<= 2;
	RLF	buff2, F
	RLF	buff2, F
;!   buff2 &= 0x3c;
	MOVLW	0x3c
	ANDWF	buff2, F
;!   *port_b = buff2;
	MOVF	buff2, W
	MOVWF	PORTB
;! }
	RETURN



;! /*-----------------------------------------------*/ 
;! void  alg_switched(void){ ; }
alg_switched:
	MOVLW	3
	MOVWF	l_swa_t
_alg_switched_0: 
	CALL	led_on
	CALL	delay_50ms
	CALL	led_off
	CALL	delay_50ms
	DECFSZ	l_swa_t, F
	GOTO	_alg_switched_0
	RETURN
	


;! /*-----------------------------------------------*/ 
;! int sw_alg_type(void) 
sw_alg_type:		
;! {
;!   return *port_a & 0x03;
	MOVF	PORTA, W
	ANDLW	0x03
;! }
	RETURN	
	
	
;! /*-----------------------------------------------*/ 
;! int sw_fault(void) 
sw_fault:	
;! {
;!   buff = *port_a;
	MOVF	PORTA, W
	MOVWF	buff
;!   buff >>= 2;
	RRF	buff, F
	RRF	buff, F
;!   return buff & 0x01;
	MOVF	buff, W
	ANDLW	0x01
	RETURN	
;! }
	


	
;! /*-----------------------------------------------*/ 
;! void  led_on(void){ ; }
;; ** cycles : 4
led_on:
	BSF	PORTA, PA_LED
	RETURN 
	
;! /*-----------------------------------------------*/ 
;! void  led_off(void){ ; }
led_off:
	BCF	PORTA, PA_LED
	RETURN 



;! /*-----------------------------------------------*/ 
	include		delay4mhz.asm



;! 
;! /*END*/

;;; End.
